"""
Compare the results of all models

@author José Antonio García-Díaz <joseantonio.garcia8@um.es>
@author Rafael Valencia-Garcia <valencia@um.es>
"""

import argparse
import csv
import json
import glob
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
import sys
import textwrap

from pathlib import Path
from datetime import datetime
from tqdm import tqdm
from dlsdatasets.DatasetResolver import DatasetResolver
from utils.Parser import DefaultParser


def main ():

    # var parser
    parser = DefaultParser (description = 'Evaluate the dataset')
    
    
    # Add features
    parser.add_argument ('--source', 
        dest = 'source', 
        default = 'test', 
        help = 'Determines the source to evaluate', 
        choices = ['all', 'train', 'test', 'val']
    )
    
    
    # Add folder name
    parser.add_argument ('--folder', dest = 'folder', default = '', help = 'Select a name of the model')
    
    
    # @var args Get arguments
    args = parser.parse_args ()
    
    
    # @var dataset_resolver DatasetResolver
    dataset_resolver = DatasetResolver ()
    
    
    # @var dataset Dataset This is the custom dataset for evaluation purposes
    dataset = dataset_resolver.get (args.dataset, args.corpus, args.task, False)
    
    
    # Set the correct dataset filename
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')


    # @var df DataFrame
    df = dataset.get ()


    # @var labels Dataframe
    labels = dataset.get_available_labels ()
    if not labels:
        labels = ['label']
    
    
    # @var float_format String To set the number of decimals
    float_format = '%.10f'
    
    
    # @var task_type
    task_type = dataset.get_task_type ()
    
    
    # @var folders_to_iterate List
    folders_to_iterate = [args.folder] if args.folder else glob.glob (dataset.get_working_dir (dataset.task, 'results', args.source, '*'))
    folders_to_iterate = [folder for folder in folders_to_iterate if folder != 'benchmarks']
    
    
    # @var pbar TQDM
    pbar = tqdm (folders_to_iterate)
    
    
    # @var df_classification_reports List
    df_classification_reports = []
    

    # Iterate over folders
    for filename in pbar:
        
        # @var folder String
        folder = os.path.basename (os.path.normpath (filename))

        
        # Indicate the folder
        pbar.set_description (folder)


        # @var training_resume_file_path String
        training_resume_file_path = dataset.get_working_dir (args.task, 'models', folder, 'training_resume.json')

        
        # Normal behaviour. The resume file exists
        if os.path.isfile (training_resume_file_path):
            with open (training_resume_file_path) as json_file:
                training_resume = json.load (json_file)
        
        else:
            print ('Skipping {}'.format (training_resume_file_path))
            continue
                
                
        # @var classification_report_path String
        classification_report_path = os.path.join (filename, 'classification_report.pickle')
        
        
        # Skip
        if not os.path.isfile (classification_report_path):
            print ('Skipping {}'.format (classification_report_path))
            continue

        
        # @var df_classification_report DataFrame
        df_classification_report = pd.read_pickle (classification_report_path)
        df_classification_report['folder'] = folder
        df_classification_report['model'] = training_resume['model']
        
        
        # Append to the list of classification report
        df_classification_reports.append (df_classification_report)
        
        
        
    # @var df_reports DataFrame Multiindex
    df_reports = pd.concat (df_classification_reports).reset_index ().set_index (['index', 'folder', 'model'])


        
    if 'regression' == task_type:
    
        # @var title String
        title = dataset.get_title ()
        
        
        # @var subtitle String
        subtitle = 'Label'


        # @var html_string String
        html_string = '''
        <!DOCTYPE html>
        <html lang="en">
            <head><title>{title} {subtitle}</title></head>
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/purecss@2.1.0/build/pure-min.css" integrity="sha384-yHIFVG6ClnONEA5yB5DJXfW2/KC173DIQrYoZMEtBvGzmf0PKiGyNEqe9N6BNDBH" crossorigin="anonymous">
            <link rel="icon" type="image/x-icon" href="../../../../favicon.ico">
            
            <style>
                main {{
                    margin: 0 auto; 
                    max-width: 80rem; 
                    text-align: center
                }}
                
                main table {{
                    margin: 1rem auto;
                }}
                
                footer {{
                    position: fixed;
                    bottom: 0;
                    left: 0;
                    right: 0;
                    background: #333;
                    color: white;
                    padding: 0.3rem 0.7rem;
                    font-size: .75rem;
                }}
                
                tbody td {{
                    font-family: monospace;
                }}
            </style>
            <body>
                <main>
                    <div class="pure-g">
                        <div class="pure-u-1-1">
                            <h1>{title} <small>{subtitle}</small></h1>
                            {table}
                        </div>
                    </div>
                </main>
                <footer>
                    Generated at: {current_time}
                </footer>
            </body>
        </html>
        '''
        print (df_reports)
    
        table = df_reports.reset_index (level = ['model', 'index'], drop = True).fillna ('').to_html (
            classes = 'pure-table pure-table-bordered',
            float_format = float_format
        ).replace ('<td>', '<td align="right">')
    
        
        # Store
        with open (dataset.get_working_dir (dataset.task, 'results', args.source, 'benchmarks', 'label.html'), 'w') as f:
            f.write (html_string.format (
                title = title,
                subtitle = subtitle,
                table = table,
                current_time = datetime.today ().strftime ('%Y-%m-%d %H:%M:%S')
            ))

    
        return

    # @var labels_to_report List
    labels_to_report = labels + ['macro avg', 'weighted avg', 'accuracy']


    # Set style
    plt.style.use ('ggplot')
    plt.tight_layout ()


    # Horizontal subplots
    fig = plt.figure (constrained_layout = True, figsize = (30, 15))
    fig.suptitle ('Benchmark of the different models with the ' + args.source + ' split', fontsize = '32')
    
    
    # Create two subfigures, the first one for reporting the labels and the other 
    # to report the main metrics
    subfigs = fig.subfigures (nrows = 2, ncols = 1)


    # @var palette Iterator
    palette = iter (sns.husl_palette (len (labels_to_report)))
    
    
    # Iterate over the sub-figures
    for row, subfig in enumerate (subfigs):
    
        # @var is_label_row boolean
        is_label_row = row == 0;
    
    
        # Main title
        subfig.suptitle ('By label' if is_label_row else 'Global results', fontsize = '24')
    
    
        # Create subplots
        subfig._subplots = subfig.subplots (nrows = 1, ncols = len (labels) if is_label_row else 3, sharey = True)
        for subplot in subfig._subplots:
            subplot.set_xlim (0, 100)
    
    
    # @var subfig_x_index int
    subfig_x_index = 0
    
    
    # @var subfig_y_index int
    subfig_y_index = 0



    # Get the best model per metric
    for metric in labels_to_report:
        
        # @var df_reports_by_metric DataFrame Get the results grouped by metric
        df_reports_by_metric = df_reports.loc[[metric]].reset_index (level = 'index', drop = True)
        df_reports_by_metric = df_reports_by_metric.sort_index ()
        
        
        # @var title String
        title = dataset.get_title ()
        
        
        # @var subtitle String
        subtitle = metric


        # @var html_string String
        html_string = '''
        <!DOCTYPE html>
        <html lang="en">
            <head><title>{title} {subtitle}</title></head>
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/purecss@2.1.0/build/pure-min.css" integrity="sha384-yHIFVG6ClnONEA5yB5DJXfW2/KC173DIQrYoZMEtBvGzmf0PKiGyNEqe9N6BNDBH" crossorigin="anonymous">
            <link rel="icon" type="image/x-icon" href="../../../../favicon.ico">
            
            <style>
                main {{
                    margin: 0 auto; 
                    max-width: 80rem; 
                    text-align: center
                }}
                
                main table {{
                    margin: 1rem auto;
                }}
                
                footer {{
                    position: fixed;
                    bottom: 0;
                    left: 0;
                    right: 0;
                    background: #333;
                    color: white;
                    padding: 0.3rem 0.7rem;
                    font-size: .75rem;
                }}
                
                tbody td {{
                    font-family: monospace;
                }}
            </style>
            <body>
                <main>
                    <div class="pure-g">
                        <div class="pure-u-1-1">
                            <h1>{title} <small>{subtitle}</small></h1>
                            {table}
                        </div>
                    </div>
                </main>
                <footer>
                    Generated at: {current_time}
                </footer>
            </body>
        </html>
        '''
        
        # Update metrics
        df_reports_by_metric = df_reports_by_metric.reset_index (level = ['model'])
        
        df_reports_by_metric.index = df_reports_by_metric.index.to_series ().str.replace (r'^deep-learning-(\w)+-.+', r'KI', regex = True)
        df_reports_by_metric.index = df_reports_by_metric.index.to_series ().str.replace (r'^deep-learning-','', regex = True)
        df_reports_by_metric.index = df_reports_by_metric.index.to_series ().str.replace (r'^ensemble-(\w+)', r'EL (\1)', regex = True)
        df_reports_by_metric.index = df_reports_by_metric.index.str.upper ()

        
        # Sort and remove model
        df_reports_by_metric = df_reports_by_metric.sort_values (
            by = ['model', 'folder'], 
            ascending = [True, True]
        )
        
        df_reports_by_metric = df_reports_by_metric.drop (columns = ['model']) 

        
        # @var table String
        table = df_reports_by_metric.drop (['support'], axis = 1).fillna ('').to_html (
            classes = 'pure-table pure-table-bordered',
            float_format = float_format
        ).replace ('<td>', '<td align="right">')
        
        
        # Store
        with open (dataset.get_working_dir (dataset.task, 'results', args.source, 'benchmarks', metric.replace (' ', '-') + '.html'), 'w') as f:
            f.write (html_string.format (
                title = title,
                subtitle = subtitle,
                table = table,
                current_time = datetime.today ().strftime ('%Y-%m-%d %H:%M:%S')
            ))
        
        
        # Latex
        # Store a copy in latex
        # @var report_latex DataFrame
        report_latex = df_reports_by_metric.copy ()
        report_latex = report_latex.drop ('support', axis = 1)
        
        
        # Update fields to adjust precision
        # Beside,s we highlight the best result per column
        for field in ['precision', 'recall', 'f1-score']:

            # @var max_value float
            max_value = report_latex[field].max ()


            # @var max_indexes List Get all row indexes where the maximum value appear
            max_indexes = report_latex.loc[report_latex[field] == max_value, field].T.index.tolist ()


            # Update all the column to format number
            report_latex[field] = r'\no{' + report_latex[field].astype (str) + '}'


            # Update report
            report_latex.loc[max_indexes, field] = r'\bo{' + str (max_value) + '}'
            
            
            # Update the specific field
            report_latex.loc[max_indexes, field] = report_latex.loc[max_indexes, field].replace (r'no{', r'bo{')
        
        
        # @var latex_string String
        latex_string = '''\
\\begin{{table}}[hbp!]
\\npdecimalsign{{.}}
\\nprounddigits{{3}}
\\npthousandsep{{}}
\\centering
\\caption{{{title}}}\\label{{tab:{label}}}
{table}
\\end{{table}}'''


        # @var table String Get the inner report
        table = report_latex.fillna ('').to_latex (
            column_format = 'lrrr',
            index = True, 
            escape = False,
            float_format = float_format
        )


        label = ['results', dataset.dataset, dataset.corpus, dataset.task, args.folder, args.source]
        label = [x for x in label if x]
        label = '-'.join (label)
        
        
        # Store the LATEX report
        with open (dataset.get_working_dir (dataset.task, 'results', args.source, 'benchmarks', metric.replace (' ', '-') + '.latex'), 'w') as f:
            f.write (latex_string.format (
                title = title,
                subtitle = subtitle,
                label = label,
                table = table
            ))
        
        
        # @var x_axis List
        x_axis = df_reports_by_metric.index.tolist ()
        x_axis = [item.replace ('deep-learning-', 'DL-').replace ('ensemble-', 'EL-').upper () for item in x_axis]

        
        # @var columns List
        columns = ["\n".join (textwrap.wrap (i, len (max (x_axis, key = len)))) for i in x_axis]
        

        # Set new column
        df_reports_by_metric['y_axis'] = columns
        
        
        # @var y_axis int
        if not metric in labels and subfig_y_index == 0:
            subfig_y_index = 1
            subfig_x_index = 0

        
        # @var ax subplot
        ax = sns.barplot (
            ax = subfigs[subfig_y_index]._subplots[subfig_x_index], 
            x = 'f1-score', 
            y = 'y_axis', 
            color = next (palette), 
            data = df_reports_by_metric
        )
        
        
        # @var support String
        support = str (int (df_reports_by_metric.iloc[0]['support']))
        
        
        # Set the X-axis label
        ax.set_xlabel (metric + " (" + support + ")", fontsize = 16)
        
        
        # Set the value on the bar
        ax.bar_label (ax.containers[0], label_type = 'edge', fmt='%.1f')
        
        
        # Set the Y-ticks to report the name of the models
        ax.set_yticklabels (columns, rotation = 0, ha = 'right', rotation_mode = 'anchor')
        
        
        # Place a vertical line, to highlight the best result
        ax.axvline (df_reports_by_metric['f1-score'].max (), linestyle = '--', color = 'red')
        ax.axvline (df_reports_by_metric['f1-score'].mean (), linestyle = '--', color = 'blue')
        
        
        # Increment x_axis
        subfig_x_index = subfig_x_index + 1
        
        
    # @var model_comparison_pdf String
    model_comparison_pdf = dataset.get_working_dir (dataset.task, 'results', args.source, 'benchmarks', 'labels.pdf')
        
        
    # Save model comparative
    plt.savefig (model_comparison_pdf, bbox_inches = 'tight')
    plt.clf ()



if __name__ == "__main__":
    main ()
