"""
    Show dataset statistics for label and split distribution
    
    @author José Antonio García-Díaz <joseantonio.garcia8@um.es>
    @author Rafael Valencia-Garcia <valencia@um.es>
"""

import sys
import os.path
import pandas as pd
import numpy as np
import seaborn as sns

from datetime import datetime
from dlsdatasets.DatasetResolver import DatasetResolver
from utils.Parser import DefaultParser


def main ():
    
    # var parser
    parser = DefaultParser (description = 'Generate statistics of the dataset')
    

    # @var args Get arguments
    args = parser.parse_args ()
    
    
    # @var resolver Resolver
    resolver = DatasetResolver ()
    
    
    # @var dataset Dataset
    dataset = resolver.get (args.dataset, args.corpus, args.task, args.force)
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')
    
    
    # @var df Dataframe The dataframe
    df = dataset.get ()
    
    
    # @var task_type String
    task_type = dataset.get_task_type ()
    
    
    # @var split_field String The split field to get the splits
    split_field = '__split'
    
    
    # @var label_field String The default label field
    label_field = 'label'
    
    
    # @var grouping_per_user Boolean 
    # Determines if the statistics of the dataset 
    # should be grouped by tweet (default) or by group
    grouping_per_user = dataset.group_results_by_user ()
    
    
    # Keep train, validation, and test only
    df = df.loc[~df[split_field].isna ()]
    df = df.loc[df[split_field].isin (['train', 'val', 'test'])]
    
    
    # To avoid hide information
    with pd.option_context ('display.max_rows', 1000, 'display.max_columns', 10):
        
        # Information concerning the dates
        if 'twitter_created_at' in df.columns:

            # @var df_dates DataFrame Force date field is a date
            df_dates = df.copy ()
            df_dates['twitter_created_at'] = pd.to_datetime (df_dates['twitter_created_at'], errors = 'coerce')


            # @var df_dates_by_month DataFrame 
            df_dates_by_month = df_dates['twitter_created_at'].groupby (df_dates['twitter_created_at'].dt.to_period ('M')).agg ('count')
            
            
            # @var min_date String
            min_date = df_dates.set_index ('twitter_created_at').index.min ()
            
            
            # @var max_date String
            max_date = df_dates.set_index ('twitter_created_at').index.max ()
            

            print ()
            print ('dataset dates')
            print ('-------------')
            
            print ()
            print ('min and max dates')
            print (min_date)
            print (max_date)
            
            print ()
            print ('distribution')
            print (df_dates_by_month)


            # @var dates_by_month_path String
            dates_by_month_path = dataset.get_working_dir (dataset.task, 'statistics', 'date-distribution.csv')

            
            # Save in disk
            df_dates_by_month.to_csv (dates_by_month_path, index = False)


        # Information concerning user distribution
        if 'user' in df.columns:
            print ()
            print ('user distribution')
            print ('-----------------')
            
            # @var user_counts Series
            user_counts = df['user'].value_counts (sort = False)
            
            
            # @var how_many_different_users int
            how_many_different_users = len (user_counts)
            
            
            # If there are multiple users, we get some useful statistics
            if how_many_different_users >= 10:
                print ('The mean of documents per user is {}'.format (user_counts.mean ()))
                print ('The median of documents per user is {}'.format (user_counts.median ()))
                print ('The mode of documents per user is {}'.format (user_counts.mode ()))
                print ('The standard deviation of documents per user is {}'.format (user_counts.std ()))
            
            else:
                print (user_counts.sort_index ())
                print (df['user'].value_counts (normalize = True, sort = False).sort_index ())
        
        
        # Data distribution
        # @var split_df DataFrame
        split_df = pd.DataFrame ()
        split_df = split_df.assign (train = '')
        split_df = split_df.assign (val = '')
        split_df = split_df.assign (test = '')
        
        
        # @var splits List
        splits = ['train', 'val', 'test']
        

        # Get the stastics per split
        # Multiclassification
        if task_type == 'classification':
            for split in splits:
                split_df[split] = dataset.get_split (df, split, split_field = split_field)[label_field].value_counts (sort = False).sort_index ()
        
        # Multilabel
        elif task_type == 'multi_label':
            for split in splits:
                
                # @var df_distribution DataFrame
                df_distribution = dataset.get_split (df, split, split_field = split_field)
                
                
                # @var df_labels DataFrame
                df_labels = df_distribution[label_field].str.split ('; ', expand=True)
                
                
                # @var columns List
                columns = [df_labels[column] for column in df_labels]
                
                
                # Update labels
                df_labels = pd.concat (columns, axis = 0, ignore_index = True).dropna (axis = 'rows')
                df_distribution = df_labels.value_counts (sort = False).sort_index ()
                split_df[split] = df_distribution

        # Regression
        elif task_type == 'regression':
            print (df.loc[df[label_field].notnull()][label_field].describe ())
            print (df.loc[df[label_field].notnull()][label_field].mode ())
            sys.exit ()


    # Grouping per user
    if grouping_per_user:
        split_df = split_df / user_counts.mean ()
        split_df = split_df.astype (int)
    
    
    # @var html_distribution_path String
    html_distribution_path = dataset.get_working_dir (dataset.task, 'statistics', 'label_distribution_per_split.html')
    
    
    # @var latex_distribution_path String
    latex_distribution_path = dataset.get_working_dir (dataset.task, 'statistics', 'label_distribution_per_split.latex')


    # @var chart_distribution_path String
    chart_distribution_path = dataset.get_working_dir (dataset.task, 'statistics', 'label_distribution_per_label.pdf')


    # Add a total per label
    split_df['total'] = split_df.sum (axis = 1)


    # Defining colors for the pie chart
    if task_type in ['classification', 'multi_label']:
        
        # @var colors List
        colors = dataset.get_colors ()


        # @var available_labels List
        available_labels = dataset.get_available_labels ()


        # @var explode Tuple To space the pie chart
        explode = (0.05,) * len (available_labels)


        # Get the plot
        plot = split_df.plot (
            kind = 'pie', 
            y = 'total', 
            autopct = '%1.0f%%',
            figsize = (10, 10),
            explode = explode,
            colors = colors
        )
        fig = plot.get_figure ()
        fig.savefig (chart_distribution_path)


    # Add a total per row
    split_df.loc['total'] = split_df.sum (numeric_only = True, axis = 0)

    
    # Remove floats
    split_df = split_df.fillna (-1)
    split_df = split_df.astype (int)
    split_df = split_df.astype (str)
    split_df = split_df.replace ('-1', np.nan)
    
    
    # @var latex_string String
    latex_string = '''\
\\begin{{table}}[hbp!]
\\centering
\\caption{{{title}}}\\label{{tab:{label}}}
{table}
\\end{{table}}'''


    # @var dataset.dataset String
    title = dataset.get_title () 
    
    
    # @var subtitle String
    subtitle = dataset.corpus + ' ' + dataset.task 
    
    
    # @var table String Get the inner report
    table = split_df.fillna ('').to_latex (
        column_format = 'llrrrr',
        index = True, 
        escape = False,
        float_format = '%.0f'
    )
    
    
    # Store the LATEX report
    with open (latex_distribution_path, 'w') as f:
        f.write (latex_string.format (
            title = title,
            subtitle = subtitle,
            table = table,
            label = 'dataset-statistics-' + args.corpus
        ))
    

    # Print useful data
    print (split_df)
    
    
    # @var html_string String
    html_string = '''
    <!DOCTYPE html>
    <html lang="en">
        <head><title>{title} {subtitle}</title></head>
        <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/purecss@2.1.0/build/pure-min.css" integrity="sha384-yHIFVG6ClnONEA5yB5DJXfW2/KC173DIQrYoZMEtBvGzmf0PKiGyNEqe9N6BNDBH" crossorigin="anonymous">
        <link rel="icon" type="image/x-icon" href="../../../favicon.ico">
        <style>
            main {{
                margin: 0 auto; 
                max-width: 80rem; 
                text-align: center
            }}
            
            main table {{
                margin: 1rem auto;
            }}
            
            footer {{
                position: fixed;
                bottom: 0;
                left: 0;
                right: 0;
                background: #333;
                color: white;
                padding: 0.3rem 0.7rem;
                font-size: .75rem;
            }}
        </style>
        <body>
            <main>
                <div class="pure-g">
                    <div class="pure-u-1-1">
                        <h1>{title} <small>{subtitle}</small></h1>
                        {table}
                    </div>
                </div>
            </main>
            <footer>
                Generated at: {current_time}
            </footer>
        </body>
    </html>
    '''
    
    
    # @var table String
    table = split_df.fillna ('').to_html (classes = 'pure-table pure-table-bordered').replace ('<td>', '<td align="right">')


    # Store
    with open (html_distribution_path, 'w') as f:
        f.write (html_string.format (
            title = title,
            subtitle = subtitle,
            table = table,
            current_time = datetime.today ().strftime ('%Y-%m-%d %H:%M:%S')
        ))
    

if __name__ == "__main__":
    main ()