"""
Evaluate the dataset

@author José Antonio García-Díaz <joseantonio.garcia8@um.es>
@author Rafael Valencia-Garcia <valencia@um.es>
"""

import os
import sys
import argparse
import pandas as pd
import numpy as np
import csv
import json
import sklearn
import itertools
import math
import warnings
import glob

from pathlib import Path
from functools import partial

from tqdm import tqdm
from datetime import datetime
from dlsdatasets.DatasetResolver import DatasetResolver
from dlsmodels.ModelResolver import ModelResolver
from utils.Parser import DefaultParser
from sklearn.metrics import confusion_matrix
from sklearn.metrics import multilabel_confusion_matrix
from sklearn.metrics import classification_report
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import explained_variance_score
from sklearn.metrics import mean_squared_log_error
from sklearn.metrics import median_absolute_error
from sklearn.metrics import r2_score

from sklearn.metrics import ConfusionMatrixDisplay

import seaborn as sn
import matplotlib.pyplot as plt


def main ():

    # var parser
    parser = DefaultParser (description = 'Evaluate the dataset')

    
    # Add optional architecture to evaluate
    parser.add_argument ('--architecture', 
        dest = 'architecture', 
        default = '', 
        help = 'Determines the architecture to evaluate', 
        choices = ['', 'dense', 'cnn', 'bigru', 'gru', 'lstm', 'bilstm']
    )
    
    
    # Add features
    parser.add_argument ('--source', 
        dest = 'source', 
        default = 'test', 
        help = 'Determines the source to evaluate', 
        choices = ['all', 'train', 'test', 'val']
    )
    
    
    # Add folder name
    parser.add_argument ('--folder', dest = 'folder', default = '', help = 'Select a name of the model')
    
    
    # @var args Get arguments
    args = parser.parse_args ()


    # @var model_resolver ModelResolver
    model_resolver = ModelResolver ()
    
    
    # @var dataset_resolver DatasetResolver
    dataset_resolver = DatasetResolver ()
        
    
    # @var architecture String
    architecture = args.architecture
    
    
    # @var dataset Dataset This is the custom dataset for evaluation purposes
    dataset = dataset_resolver.get (args.dataset, args.corpus, args.task, False)
    
    
    # Set the correct dataset filename
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')
    
    
    # @var folders_to_iterate List
    folders_to_iterate = [args.folder] if args.folder else glob.glob (dataset.get_working_dir (dataset.task, 'models', '*'))


    # @var should_i_show_reports boolean
    should_i_show_reports = len (folders_to_iterate) == 1


    # @var pbar TQDM
    pbar = tqdm (folders_to_iterate)


    # Iterate over folders
    for filename in pbar:
    
        if 'benchmark' in filename:
            continue
    
        
        # @var folder String
        folder = os.path.basename (os.path.normpath (filename))

        
        # Indicate the folder
        pbar.set_description (folder)


        # @var training_resume_file_path String
        training_resume_file_path = dataset.get_working_dir (dataset.task, 'models', folder, 'training_resume.json')


        # Normal behaviour. The resume file exists
        if os.path.isfile (training_resume_file_path):
            with open (training_resume_file_path) as json_file:
                training_resume = json.load (json_file)

        # For keeping compatibility
        else:
            raise ValueError ('Folder does not contains a valid resume file', training_resume_file_path) 
            sys.exit ()


        # Avoid models based on transformers (by the moment)
        if 'transformers' == training_resume['model']:
            print ("Skipping {}".format (training_resume['model']))
            continue

        
        # @var model Model
        model = model_resolver.get_from_resume (training_resume)
        
        
        # Attach the dataset to the model
        model.set_dataset (dataset)
        model.is_merged (dataset.is_merged)

        
        # Define the criterion for selecting the best model
        # based on the architecture
        if architecture:
            model.set_best_model_criteria ({
                'architecture': architecture
            })
        
        
        # @var df_full DataFrame
        df_full = dataset.get ()

    
        # Replace the dataset to contain only the test or val-set
        if args.source in ['train', 'val', 'test']:
            dataset.default_split = args.source


        # @var df DataFrame Get from the default split
        df = dataset.get ()

    
        # Predict this feature set
        model.predict (callback = partial (callback, 
            dataset = dataset,
            model = model, 
            args = args,
            df = df,
            training_resume = training_resume,
            should_i_show_reports = should_i_show_reports
        ))
        
        
        # Clear session
        model.clear_session ();



def callback (feature_key, y_pred, model_metadata, args, dataset, model, training_resume, should_i_show_reports, df):
    """
    This is the callback for when we get the predictions
    
    @var feature_key String
    @var y_pred List
    @var model_metadata Dict
    """
    
    # @var task_type String
    task_type = dataset.get_task_type ()
    
    
    # @var dataset Dataset
    dataset = model.dataset


    # @var y_real List|Matrix
    y_real = dataset.get_vector_labels (df)
    
    
    # @var labels List
    labels = dataset.get_available_labels ()


    # @var float_format String To set the number of decimals
    float_format = '%.10f'


    # @var dataset.dataset String The title of the dataset
    title = dataset.get_title () 
    
    
    # @var subtitle String
    subtitle = dataset.corpus + ' ' + dataset.task 
    
    
    # @var model_name String
    model_name = model.get_folder ()


    # @var paths Dict
    paths = {
        'probabilities': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'probabilities.csv'),
        'classification_report_html': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'classification_report.html'),
        'predictions': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'y_pred.csv'),
        'dataframe': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'classification_report.pickle'),
        'classification_report_latex': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'classification_report.latex'),
        'confusion_matrix_latex': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'confusion_matrix.latex'),
        'confusion_matrix_heatmap': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'confusion_matrix.pdf'),
        'weights': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'weights.csv'),
        'predictions': dataset.get_working_dir (dataset.task, 'results', args.source, model_name, args.architecture, 'predictions.csv')
    }
    
    
    # @var y_real_labels_available boolean
    y_real_labels_available = not pd.isnull (y_real).all ()
    if 'regression' == task_type:
        y_real_labels_available = None
    

    # @var group_results_by_user Boolean
    group_results_by_user = dataset.group_results_by_user ()
    
    
    # Readapt DF
    if group_results_by_user:

        # Attach predictions
        df['y_pred'] = y_pred


        # Average probabilities by user
        if 'probabilities' in model_metadata:

            # Adjust data
            model_metadata['probabilities'] = [item[:] for item in model_metadata['probabilities']]
            df = df.assign (probabilities = model_metadata['probabilities'])

            model_metadata['probabilities'] = df \
                .groupby ('user', as_index = False)['probabilities'] \
                .apply (lambda test: np.mean (list (test.values), axis = 0)).tolist ()


        # Readapt DF
        df = df.drop_duplicates ('user', keep = 'first')
        
        
        # @var df_group
        df_group = df.groupby (by = 'user', sort = False)
        
        
        # Reassign preds by user, if necessary
        y_pred = df_group['y_pred'].apply (lambda x: x.mode ().iloc[0]).tolist ()
        y_real = df_group['label'].apply (lambda x: x.mode ().iloc[0]).tolist ()
    
    # Generate reports when the truth is available
    if y_real_labels_available:
    
        # Classification reports
        # =========================================================
        # @var report DataFrame|None
        report = pd.DataFrame (classification_report (
            y_true = y_real, 
            y_pred = y_pred, 
            digits = 5,
            output_dict = True,
            target_names = labels
        )).T
    
    
        # Adjust the report in a scale 0-100
        for label in ['precision', 'recall', 'f1-score']:
            report[label] = report[label].mul (100)


        # Calculate
        if should_i_show_reports:
            print ()
            print ("classification report")
            print ("---------------------")
            print (report.to_markdown ())

        
        # Save the classification report in disk
        report.to_pickle (paths['dataframe'])
        
        
        # Store a copy in latex
        report_latex = report.copy ()
        report_latex = report_latex.drop ('support', axis = 1)
        
        
        # Update fields to adjust precision
        for field in ['precision', 'recall', 'f1-score']:
            report_latex[field] = '\no{' + report_latex[field].astype (str) + '}'
        
        
        # @var latex_string String
        latex_string = '''\
\\begin{{table}}[hbp!]
\\npdecimalsign{{.}}
\\nprounddigits{{3}}
\\npthousandsep{{}}
\\centering
\\caption{{{title}}}\\label{{tab:results-{source}-{folder}}}
{table}
\\end{{table}}'''


        # @var table String Get the inner report
        table = report_latex.fillna ('').to_latex (
            column_format = 'lrrr',
            index = True, 
            escape = False,
            float_format = float_format
        )
        
        
        # Store the LATEX report
        with open (paths['classification_report_latex'], 'w') as f:
            f.write (latex_string.format (
                title = title,
                subtitle = subtitle,
                table = table,
                folder = args.folder,
                source = args.source
            ))
        
        
        # @var html_string String
        html_string = '''
        <!DOCTYPE html>
        <html lang="en">
            <head><title>{title} {subtitle}</title></head>
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/purecss@2.1.0/build/pure-min.css" integrity="sha384-yHIFVG6ClnONEA5yB5DJXfW2/KC173DIQrYoZMEtBvGzmf0PKiGyNEqe9N6BNDBH" crossorigin="anonymous">
            <link rel="icon" type="image/x-icon" href="../../../../../favicon.ico">
            
            <style>
                main {{
                    margin: 0 auto; 
                    max-width: 80rem; 
                    text-align: center
                }}
                
                main table {{
                    margin: 1rem auto;
                }}
                
                footer {{
                    position: fixed;
                    bottom: 0;
                    left: 0;
                    right: 0;
                    background: #333;
                    color: white;
                    padding: 0.3rem 0.7rem;
                    font-size: .75rem;
                }}
                
                tbody td {{
                    font-family: monospace;
                }}
                
                
                tbody tr:nth-last-child(-n+{rows}) th,
                tbody tr:nth-last-child(-n+{rows}) td {{
                    color: #333;
                    background: #F3EED9;
                    font-weight: bold;
                }}
                
                dl {{
                
                }}
                
                dt, dd {{
                    display: inline-block;
                    vertical-align: top;
                }}
                
                dt {{
                    font-weight: bold;
                }}
                
                dd {{
                    margin: 0 3rem 0 0;
                }}
            </style>
            <body>
                <main>
                    <div class="pure-g">
                        <div class="pure-u-1-1">
                            <h1>{title} <small>{subtitle}</small></h1>
                            {table}
                        </div>
                    </div>
                    
                    <dl>
                        <dt>Task type</dt> <dd>{task_type}</dd>
                        <dt>Scoring metric</dt> <dd>{scoring_metric}</dd>
                        <dt>Folder</dt> <dd>{folder}</dd>
                        <dt>Source</dt> <dd>{source}</dd>
                    </dl>
                    
                </main>
                <footer>
                    Generated at: {current_time}
                </footer>
            </body>
        </html>
        '''
        
        
        # Change type of support table
        report['support'] = report['support'].astype (int)
        
        
        # @var scoring_metric String
        scoring_metric = dataset.get_scoring_metric ()
        
        
        # @var table String
        table = report.fillna ('').to_html (
            classes = 'pure-table pure-table-bordered',
            float_format = float_format
        ).replace ('<td>', '<td align="right">')


        # Store
        with open (paths['classification_report_html'], 'w') as f:
            f.write (html_string.format (
                title = title,
                subtitle = subtitle,
                table = table,
                task_type = task_type,
                scoring_metric = scoring_metric,
                folder = args.folder,
                source = args.source,
                rows = 3 if 'classification' == task_type else 4,
                current_time = datetime.today ().strftime ('%Y-%m-%d %H:%M:%S')
            ))
            
            
        # Save predictions
        if 'classification' == task_type:
            
            if str (y_pred[0]).isdigit ():
                y_pred_df = pd.DataFrame (
                    np.array (y_pred).astype (np.int), 
                    columns = labels, 
                    index = df.index
                )
            else:
                y_pred_df = pd.DataFrame (
                    np.array (y_pred), 
                    columns = ['y_pred'], 
                    index = df.index
                )
                    
        elif 'multi_label' == task_type:
            y_pred_df = pd.DataFrame (np.array (y_pred), columns = labels, index = df.index)
        
        elif 'regression' == task_type:
            y_pred_df = pd.DataFrame (y_pred, index = df.index)
        
        y_pred_df.to_csv (paths['predictions'])
            
            
        # Multilabel accuracy
        if should_i_show_reports and 'multi_label' == task_type:
            print ()
            print ('precision_recall_fscore_support (micro)')
            print ("--------------------------------------")
            print (precision_recall_fscore_support (
                y_true = y_real, 
                y_pred = y_pred, 
                average = 'micro', 
                labels = labels if task_type == 'classification' else None)
            )
    
        
        # Confusion matrices
        # @var labels_in_cf List
        labels_in_cf = [label.replace ('-', ' ') for label in labels]
        labels_in_cf = [(label[:12] + '...') if len (label) > 12 else label for label in labels_in_cf]

        
        # Confusion matrix for classification problems
        if 'classification' == task_type:
            
            # @var cm ndarray
            cm = sklearn.metrics.confusion_matrix (y_true = y_real, y_pred = y_pred, labels = labels)
            
            
            # Create plot
            plt.clf ()
            ax = plt.subplot ()
            sn.set (font_scale = 1)
            
            
            # Create heatmap
            heatmap = sn.heatmap (cm, annot = True, fmt = '.0f', annot_kws = {'size': 'small'}, cbar = None)
            
            
            # Create the labels
            if (len (labels)) <= 10:
                ax.xaxis.set_ticklabels (labels_in_cf, rotation = 90); 
                ax.yaxis.set_ticklabels (labels_in_cf, rotation = 30);
                for t in ax.texts: t.set_text (t.get_text())
            
            else:
                for t in ax.texts: t.set_text ('')
            
            
            # Set the axis
            ax.set (xlabel = 'Predicted', ylabel = 'Actual')


            # @var cm_df DataFrame
            cm_df = pd.DataFrame (cm, columns = labels, index = labels)

            
            if should_i_show_reports:
                print ()
                print ("confusion matrix")
                print ("-----------------")
                print (cm_df.to_markdown ())
            

            # Save confusion matrix
            cm_df.to_latex (paths['confusion_matrix_latex'], index = True)
        
            
            # Save heatmap
            plt.savefig (paths['confusion_matrix_heatmap'], bbox_inches = 'tight')
            
            
        elif 'multi_label' == task_type:
            
            # @var cm Confusion Matrix with percentages
            cm = sklearn.metrics.multilabel_confusion_matrix (
                y_true = np.array (y_real).astype (np.float), 
                y_pred = np.array (y_pred).astype (np.float)
            )
            
            
            # @var int num_labels Number of labels
            num_labels = len (labels)
            
            
            # @var int plots_per_row Number of plots per row
            ncols = 5
            
            
            # Adjust
            if num_labels % 3 == 0:
                ncols = 3
            elif num_labels % 2 == 0:
                ncols = 2
                
                
            # @var int nrows Number of rows
            nrows = int (math.ceil (len (labels) / ncols))
            
            
            # @link https://stackoverflow.com/questions/62722416/plot-confusion-matrix-for-multilabel-classifcation-python
            f, axes = plt.subplots (nrows, ncols, figsize = (25, len (labels)))
            
            
            # @var axex
            axes = axes.ravel ()
            
            
            # Create confusion matrices
            for i in range (len (labels)):
                disp = ConfusionMatrixDisplay (confusion_matrix (y_real[:, i], y_pred[:, i]), display_labels = [0, i])
                disp.plot (ax = axes[i], values_format = '.4g')
                disp.ax_.set_title (f'{labels[i]}')
                if not (i % ncols == 0):
                    disp.ax_.set_xlabel ('')
                    
                disp.im_.colorbar.remove ()
            
            plt.subplots_adjust (wspace = 0.10, hspace = 0.1)
            f.colorbar (disp.im_, ax = axes)
            plt.savefig (paths['confusion_matrix_heatmap'], bbox_inches = 'tight')

    
    # Check if does it make sense to store the probabilities
    if 'probabilities' in model_metadata and model_metadata['probabilities'] is not None:

        # @var do_i_have_one_label boolean To check if the probabilities encoded has one or two labels
        do_i_have_one_label = len (model_metadata['probabilities'][0]) == 1


        if task_type in ['multi_label', 'classification']:
        
            # Attach the labels
            probabilities_df = pd.DataFrame (model_metadata['probabilities'], 
                columns = [labels[0]] if do_i_have_one_label else labels
            )
            
            
            # For case of binary labels calculate the contrary possibility
            if do_i_have_one_label:
                probabilities_df[labels[1]] = 1 - model_metadata['probabilities']
        
        elif 'regression' == task_type:
            probabilities_df = pd.DataFrame (model_metadata['probabilities'], 
                columns = ['probability']
            )
        
        
        # Include the real label and the feature set employed
        probabilities_df = probabilities_df \
            .assign (label = y_real) \
            .assign (features = feature_key)
        
        
        # Reorder
        probabilities_df = probabilities_df[['features'] + [col for col in probabilities_df.columns if col != 'features']]
    
    
        # Save probabilities
        probabilities_df.to_csv (paths['probabilities'], index = False)
    
    
    # Store weights
    # ---------------
    if 'weights' in model_metadata:
        pd.DataFrame (model_metadata['weights'], index=[0]).to_csv (paths['weights'], index = False)
    
    
    
    # Store predictions
    # -----------------
    # @var df_split Ensure if we already had the data processed
    df_split = dataset.get ()
    
    
    # Save predictions
    if 'classification' == task_type:
        pd.DataFrame ({
            'y_pred': y_pred,
            'y_real': y_real
        }).to_csv (paths['predictions'], index = True, quoting = csv.QUOTE_ALL)


    if 'multi_label' == task_type:
        
        # @var lb  MultiLabelBinarizer
        lb = sklearn.preprocessing.MultiLabelBinarizer ()
        
        
        # Fit the multi-label binarizer
        lb.fit ([labels])
        
        
        # Create dataframe
        pd.DataFrame ({
            'y_pred': [','.join (list (x)) for x in lb.inverse_transform (y_pred)],
            'y_real': [','.join (list (x)) for x in lb.inverse_transform (y_real)]
        }).to_csv (paths['predictions'], index = True, quoting = csv.QUOTE_ALL)
    
    
    # Final metric
    if 'regression' == task_type: 
        
        report = pd.DataFrame ({
            'explained variance': explained_variance_score (y_true = y_real, y_pred = y_pred),
            'mean squared log error': mean_squared_log_error (y_true = y_real, y_pred = y_pred),
            'r2': r2_score (y_true = y_real, y_pred = y_pred),
            'mae': mean_absolute_error (y_true = y_real, y_pred = y_pred),
            'mse': mean_squared_error (y_true = y_real, y_pred = y_pred),
            'rmse': np.sqrt (mean_squared_error (y_true = y_real, y_pred = y_pred))
        }, index=[0])
        
        
        if should_i_show_reports:
            print ()
            print ("regression report")
            print ("-----------------")
            print (report.to_markdown ())
        
        # Save the classification report in disk
        report.to_pickle (paths['dataframe'])
        
        
        # Store reports in other formats
        report.to_latex (paths['classification_report_latex'], index = True, float_format = float_format)
        report.to_html (paths['classification_report_html'], float_format = float_format)


if __name__ == "__main__":
    main ()
