"""
    Generate tables
    
    @author José Antonio García-Díaz <joseantonio.garcia8@um.es>
    @author Rafael Valencia-Garcia <valencia@um.es>
"""

import os
import glob
import sys
import argparse
import pandas as pd
import numpy as np
import pickle
import json

from datetime import datetime
from tqdm import tqdm
from pathlib import Path
from dlsdatasets.DatasetResolver import DatasetResolver
from utils.Parser import DefaultParser
from dlsmodels.ModelResolver import ModelResolver

def main ():

    # var parser
    parser = DefaultParser (description = 'Generate the hyperparameter table')
    
    
    # @var args Get arguments
    args = parser.parse_args ()
    
    
    # @var dataset_resolver DatasetResolver
    dataset_resolver = DatasetResolver ()
    
    
    # @var model_resolver ModelResolver
    model_resolver = ModelResolver ()
    
    
    # @var dataset Dataset This is the custom dataset for evaluation purposes
    dataset = dataset_resolver.get (args.dataset, args.corpus, args.task, False)
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')


    # @var df Ensure if we already had the data processed
    df = dataset.get ()
    
    
    # @var models Iterator All available models
    models = glob.glob (dataset.get_working_dir (dataset.task, 'models', '*/'))
    
    
    # @var results dict
    results = {}


    # @var columns List
    columns = ['architecture', 'shape', 'num_layers', 'first_neuron', 'dropout', 'lr', 'batch_size', 'activation']
    
    
    # @var columns_type List
    columns_type = [('type/' + column) for column in columns]


    # @var pbar Iterator
    pbar = tqdm (models)

    
    # Iterate over the models
    for model in pbar:
        
        # @var directory String
        directory = model.split ('/')[-2]
        
        
        # Skip...
        if directory == '*':
            continue


        # Indicate the value
        pbar.set_description (directory)


        # @var training_resume_file String
        training_resume_file = dataset.get_working_dir (dataset.task, 'models', directory, 'training_resume.json')


        # No training resume file was found
        if not os.path.isfile (training_resume_file):
            continue

        
        # Get the details of te training
        with open (training_resume_file) as json_file:
            training_resume = json.load (json_file)
            
        
        # @var model_type String
        model_type = training_resume['model']
        
        
        # Only deep-learning models
        # @todo Improve this. Define param type
        if 'deep-learning' != model_type:
            continue
                
        
        # @var hyperparameters_filename String    
        hyperparameters_filename = dataset.get_working_dir (dataset.task, 'models', directory, 'hyperparameters.csv')


        # Skip incomplete models
        if not Path (hyperparameters_filename).is_file ():
            continue
        
        
        # @var hyperparameters_df DataFrame
        hyperparameters_df = pd.read_csv (hyperparameters_filename, index_col = False)


        # Get the best model
        hyperparameters_df = hyperparameters_df.loc[hyperparameters_df['best'] == 1].tail (1).reset_index ()
        
        
        # Append to the results
        results[directory] = hyperparameters_df[columns_type]
        

    # @var report_df Store
    report_df = pd.concat (results.values (), keys = results.keys (), ignore_index = False) \
        .rename (columns = {'type/num_layers': '\# of layers', 'first_neuron': '\# of neurons'}) \
        .drop (['type/batch_size'], axis = 1)
    report_df = report_df.sort_index ()
    report_df = report_df.droplevel (level = 1)
    
    
    # @var latex_string String
    latex_string = '''\
\\begin{{table}}[hbp!]
\\centering
\\caption{{{title}}}\\label{{tab:{label}}}
{table}
\\end{{table}}'''
    
    
    # Save latex file
    # @var table String Get the inner report
    table = report_df.fillna ('').to_latex (
        column_format = 'lllrrrrl',
        index = True, 
        escape = False,
        float_format = '%.0f'
    )
    
    
    # @var dataset.dataset String
    title = dataset.get_title () 
    
    
    # @var subtitle String
    subtitle = dataset.corpus + ' ' + dataset.task 
    
    
    # @var html_best_hyperparameters_path String
    html_best_hyperparameters_path = dataset.get_working_dir (dataset.task, 'statistics', 'best_hyperparameters.html')
        
    
    # @var latex_best_hyperparameters_path String
    latex_best_hyperparameters_path = dataset.get_working_dir (dataset.task, 'statistics', 'best_hyperparameters.latex')
    
    
    # Store the LATEX report
    with open (latex_best_hyperparameters_path, 'w') as f:
        f.write (latex_string.format (
            title = title,
            subtitle = subtitle,
            table = table,
            label = 'dataset-statistics-' + args.corpus
        ))


    # @var html_string String
    html_string = '''
    <!DOCTYPE html>
    <html lang="en">
        <head><title>{title} {subtitle}</title></head>
        <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/purecss@2.1.0/build/pure-min.css" integrity="sha384-yHIFVG6ClnONEA5yB5DJXfW2/KC173DIQrYoZMEtBvGzmf0PKiGyNEqe9N6BNDBH" crossorigin="anonymous">
        <link rel="icon" type="image/x-icon" href="../../../favicon.ico">
        <style>
            main {{
                margin: 0 auto; 
                max-width: 80rem; 
                text-align: center
            }}
            
            main table {{
                margin: 1rem auto;
            }}
            
            footer {{
                position: fixed;
                bottom: 0;
                left: 0;
                right: 0;
                background: #333;
                color: white;
                padding: 0.3rem 0.7rem;
                font-size: .75rem;
            }}
        </style>
        <body>
            <main>
                <div class="pure-g">
                    <div class="pure-u-1-1">
                        <h1>{title} <small>{subtitle}</small></h1>
                        {table}
                    </div>
                </div>
            </main>
            <footer>
                Generated at: {current_time}
            </footer>
        </body>
    </html>
    '''
    
    
    # @var table String
    table = report_df.fillna ('').to_html (classes = 'pure-table pure-table-bordered').replace ('<td>', '<td align="right">')


    # Store
    with open (html_best_hyperparameters_path, 'w') as f:
        f.write (html_string.format (
            title = title,
            subtitle = subtitle,
            table = table,
            current_time = datetime.today ().strftime ('%Y-%m-%d %H:%M:%S')
        ))

    
    print ("Best hyperparameters")
    print (report_df.to_markdown ())



if __name__ == "__main__":
    main ()
    