"""
Information Gain per class

This class calculates the Information Gain (Mutual Info) of a dataset
and uses it to select the most discrimatory features


@author José Antonio García-Díaz <joseantonio.garcia8@um.es>
@author Rafael Valencia-Garcia <valencia@um.es>
"""

import os
import sys
import argparse
import pandas as pd
import numpy as np
import pickle
import matplotlib.pyplot as plt
import seaborn as sns

from pathlib import Path
from sklearn.feature_selection import mutual_info_classif
from sklearn.feature_selection import mutual_info_regression

from dlsdatasets.DatasetResolver import DatasetResolver
from utils.Parser import DefaultParser
from features.FeatureResolver import FeatureResolver
from features.LinguisticFeaturesTransformer import LinguisticFeaturesTransformer


def main ():

    # Set theme
    sns.set_theme ()
    sns.set_palette ('pastel')


    # var parser
    parser = DefaultParser (description = 'Calculates the Information Gain (Mutual Info) per class and obtains the best LF')
    
    
    # @var args Get arguments
    args = parser.parse_args ()
    
    
    # @var dataset_resolver DatasetResolver
    dataset_resolver = DatasetResolver ()
    
    
    # @var dataset Dataset This is the custom dataset for evaluation purposes
    dataset = dataset_resolver.get (args.dataset, args.corpus, args.task, False)
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')


    # @var df Ensure if we already had the data processed
    df = dataset.get ()
    
    
    # @var task_type String
    task_type = dataset.get_task_type ()
    
    
    # @var df_train DataFrame
    df_train = dataset.get_split (df, 'train')
    df_train = df_train[df_train['label'].notna()]
    
    
    # @var feature_resolver FeatureResolver
    feature_resolver = FeatureResolver (dataset)
    
    
    # @var feature_file String
    feature_file = feature_resolver.get_suggested_cache_file ('lf', task_type)
    
    
    # @var features_cache String The file where the features are stored
    features_cache = dataset.get_working_dir (args.task, feature_file)
    
    
    # If the feautures are not found, get the default one
    if not Path (features_cache).is_file ():
        raise Exception ('features lf file are not avaiable: ' + features_cache)
        sys.exit ()
        
    
    # @var transformer Transformer
    transformer = feature_resolver.get ('lf', cache_file = features_cache)
    
    
    # @var features_df DataFrame
    features_df = transformer.transform ([])
    
    
    # @var linguistic_features List
    linguistic_features = features_df.columns.to_list ()
        
    
    # Keep only the training features
    features_df = features_df[features_df.index.isin (df_train.index)].reindex (df_train.index)
    
    
    # Attach label
    features_df = features_df.assign (label = df_train['label'])

    
    # @var unique_labels Series Bind to the label
    unique_labels = dataset.get_available_labels ()
    
    
    # @var X
    X = features_df.loc[:, features_df.columns != 'label']
    
    
    # @var y
    y = df_train['label']
    
    
    # @var mi 
    if 'classification' == task_type:
        mi = mutual_info_classif (X = X, y = y).reshape (-1, 1)
        
    elif 'regression' == task_type:
        mi = mutual_info_regression (X = X, y = y).reshape (-1, 1)
    
    
    # @var df_ig List
    df_ig = pd.DataFrame (mi, 
        columns = ['Coefficient'], 
        index = linguistic_features
    )
    
    
    # Best indexes and categories
    df_ig = df_ig.assign (feature = df_ig.index)
    
    
    # Build categories and features
    # As the names are complex (i.e. social-media-jargon-male), we prefer to get the category 
    # and the feature into two columns
    # First, we fix those categories that are two names
    df_ig['feature'] = df_ig['feature'].str.replace ('social-media', 'social media')
    
    
    # Next, we get the category as a single column
    df_ig['category'] = df_ig['feature'].str.split ('-', expand = True)[0]
    
    
    # We get the rest of the string and fed into "feature" column
    df_ig['feature'] = df_ig.dropna ()['feature'].str.split ('-', expand = True).drop (columns = 0).fillna ('').agg ('-'.join, axis=1).str.strip ('-')
    
    
    # As some features are none, we copy them from the category
    df_ig.loc[df_ig['feature'] == '', 'feature'] = df_ig.loc[df_ig['feature'] == '']['category']
    
    
    # @var acronyms Dict
    acronyms = LinguisticFeaturesTransformer.get_acronyms ()
    
    
    # Assign the acronyms
    df_ig['acronym'] = df_ig['category'].apply (lambda x: acronyms.get (x))
    df_ig = df_ig.reset_index (drop = True)


    # Create a index for ploting
    df_ig['plot_index'] = '(' + df_ig['acronym'] + ")\n" + df_ig['feature']    
    
    
    # @var labels_to_export Reorder columns
    labels_to_export = ['acronym', 'category', 'feature', 'Coefficient']

    
    # Best features for regression tasks
    if 'regression' == task_type:

        print ("by dataset")
        print ("----------")
        df_ig.index = linguistic_features
        print ("top")
        print (df_ig.sort_values (by = 'Coefficient', ascending = False).head (20).to_csv ())
        
        print ("worst")
        print (df_ig.sort_values (by = 'Coefficient', ascending = True).head (10).to_csv ())

    
    # Best features for classification tasks
    if 'classification' == task_type:
    
        # @var average_features_per_label List
        average_features_per_label = [features_df.loc[df_train.loc[features_df['label'] == label].index].mean ().to_frame ().T for label in unique_labels]
        
        
        # @var colors List
        colors = dataset.get_colors ()
        
        
        # @var df_ig_per_label DataFrame Merge features by label
        df_ig_per_label = pd \
            .concat (average_features_per_label) \
            .reset_index (drop = True) \
            .assign (label = unique_labels) \
            .set_index ('label') \
            .transpose () \
            .reset_index (drop = True)
        
        df_ig_per_label['plot_index'] = df_ig['plot_index']
        
        # Attach coefficient to the features
        df_ig_per_label = df_ig_per_label.assign (Coefficient = df_ig['Coefficient'].values)


        # Remove those features without coefficient
        df_ig_per_label = df_ig_per_label.loc[df_ig_per_label['Coefficient'] > 0]
        df_ig = df_ig.loc[df_ig['Coefficient'] > 0]
        

        # @var paths Dict
        paths = {
            'coefficient': dataset.get_working_dir (dataset.task, 'information-gain', 'by-coefficient.csv'),
            'coefficient_chart': dataset.get_working_dir (dataset.task, 'information-gain', 'by-coefficient.pdf'),
            'by_category': dataset.get_working_dir (dataset.task, 'information-gain', 'by-category.csv'),
            'label': dataset.get_working_dir (dataset.task, 'information-gain', 'by-label.csv'),
            'label_chart': dataset.get_working_dir (dataset.task, 'information-gain', dataset.dataset + '-' + dataset.corpus + '-' + dataset.task + '-by-label.pdf'),
        }


        # @var df_ig_sorted DataFrame
        df_ig_sorted = df_ig.set_index ('plot_index')[labels_to_export].sort_values (by = 'Coefficient', ascending = False)


        # Save all features ranked by coefficient
        df_ig_sorted.to_csv (paths['coefficient'], index = False)
        

        # @var plot get a plot for represeting the IG
        plot = df_ig_sorted \
            .head (10)[::-1] \
            .plot (kind = 'barh', figsize = (15, 10))


        # Labels
        plt.xlabel ('Information Gain')
        plt.ylabel ('Linguistic feature')
 
        
        # Title of plot
        plt.title ('Top 10 Information gain linguistic features')
        plt.yticks (rotation = 30)
        
        
        # @var figure Figure
        figure = plot.get_figure ()
        figure.savefig (paths['coefficient_chart'], bbox_inches = 'tight')

        
        # Clear the plot for another graph
        plt.clf () 
        
        
        # @var df_ig_per_label_sorted DataFrame
        df_ig_per_label_sorted = df_ig_per_label.sort_values (by = 'Coefficient', ascending = False)
        
        
        # Save all features by label ranked by coefficient
        df_ig_per_label_sorted[unique_labels].to_csv (paths['label'], index = False)


        # Normalize the label values. This is for observing if the 
        # linguistic feature appears more on one label or others
        df_ig_per_label_sorted = df_ig_per_label_sorted.set_index ('plot_index')[unique_labels].head (10)
        df_ig_per_label_sorted = df_ig_per_label_sorted.div (df_ig_per_label_sorted.sum (axis = 1), axis = 0)


        # Generate the plot
        plot = df_ig_per_label_sorted \
            .head (10)[::-1] \
            .plot (kind = 'barh', stacked = True, figsize = (15, 10), color = colors)


        # Labels
        plt.xlabel ('Information Gain')
        plt.ylabel ('Linguistic feature')
 
        
        # Title of plot
        plt.title ('Top 10 Information gain linguistic features')
        plt.yticks (rotation = 30)
        
        
        # @var figure Figure
        figure = plot.get_figure ()

        figure.savefig (paths['label_chart'], bbox_inches = 'tight')


        # Group features by category
        df_ig.sort_values (
            by = ['category', 'Coefficient'], 
            ascending = [True, False]
        ).groupby ('category').head (1000).to_csv (paths['by_category'], index = False)
        
        
        # By category
        for category, acronym in acronyms.items ():
            df_ig.loc[df_ig['acronym'] == acronym].sort_values (by = ['Coefficient'], ascending = False).to_csv (dataset.get_working_dir (dataset.task, 'information-gain', 'by-category-' + acronym.lower () + '.csv'), index = False)
        

if __name__ == "__main__":
    main ()
