"""
Predict and use the models

@author José Antonio García-Díaz <joseantonio.garcia8@um.es>
@author Rafael Valencia-Garcia <valencia@um.es>
"""

import os
import sys
import argparse
import pandas as pd
import numpy as np
import csv
import json
import sklearn
import itertools
import glob

from pathlib import Path

from dlsdatasets.DatasetResolver import DatasetResolver
from dlsmodels.ModelResolver import ModelResolver
from utils.Parser import DefaultParser


def main ():

    # var parser
    parser = DefaultParser (description = 'Predict using a model')

    
    # Add features
    parser.add_argument ('--source-type', 
        dest = 'source_type', 
        default = 'text',
        choices = ['text', 'csv'],
        help = 'Determine if the source is a file with texts or direct input'
    )
    
    
    # Determine the source
    parser.add_argument ('--source', 
        dest = 'source', 
        help = 'The source'
    )
    
    
    # Add folder name
    parser.add_argument ('--folder', dest = 'folder', default = '', help = 'Select a name of the model')
    
    
    # @var args Get arguments
    args = parser.parse_args ()


    # @var df_origin DataFrame
    df_origin = None


    # Data source is a CSV
    if args.source_type == 'csv':
        
        # @var dataset_path String
        dataset_path = args.source
        if not os.path.isfile (dataset_path):
            print ("The source csv does not exist")
            sys.exit ()
        
        # Read the data
        df_origin = pd.read_csv (dataset_path)
        
    
    # Datasource is text
    if args.source_type == 'text':
        df_origin = pd.DataFrame (args.source.split (','), columns = ['tweet'])


    # @var model_resolver ModelResolver
    model_resolver = ModelResolver ()
    
    
    # @var dataset_resolver DatasetResolver
    dataset_resolver = DatasetResolver ()
    
    
    # @var dataset Dataset This is the custom dataset for evaluation purposes
    dataset = dataset_resolver.get (args.dataset, args.corpus, args.task, False)
    
    
    # Set the correct dataset filename
    dataset.filename = dataset.get_working_dir (args.task, 'dataset.csv')
    
        
    # @var folder String
    folder = os.path.basename (os.path.normpath (args.folder))
    
    
    # @var training_resume_file_path String
    training_resume_file_path = dataset.get_working_dir (dataset.task, 'models', folder, 'training_resume.json')


    # Normal behaviour. The resume file exists
    if os.path.isfile (training_resume_file_path):
        with open (training_resume_file_path) as json_file:
            training_resume = json.load (json_file)

    # For keeping compatibility
    else:
        raise ValueError ('Folder does not contains a valid resume file', training_resume_file_path) 
        sys.exit ()


    # By the moment, we use only transformers
    if not 'transformers' == training_resume['model']:
        sys.exit ()

    
    # @var model Model
    model = model_resolver.get_from_resume (training_resume)
        
        
    # Attach the dataset to the model
    model.set_dataset (dataset)
    model.is_merged (dataset.is_merged)
    
    
    # @var df_full DataFrame We load the dataset to store it in memory and get 
    #                        the classification labels
    df_full = dataset.get ()
    
    
    # @var labels List In classificatio task, we need the real values
    labels = dataset.get_available_labels ()
    
    
    # Predict this feature set
    model.inference (documents = df_origin['tweet'].tolist (), callback = callback_response)
    
    
    # Clear session
    model.clear_session ();


def callback_response (y_pred):
    """
    This is the callback for when we get the predictions
    
    @var y_pred List
    """
    
    print (json.dumps  (y_pred))


if __name__ == "__main__":
    main ()
